/* 
 * File:   ExceptionHandling.h
 * Author: charlesq
 *
 * Created on November 7, 2012, 5:26 PM
 */

#ifndef EXCEPTIONHANDLING_H
#define	EXCEPTIONHANDLING_H

#include "llvm/LLVMContext.h"
#include "llvm/Constants.h"
#include "llvm/DerivedTypes.h"
#include "llvm/ExecutionEngine/ExecutionEngine.h"
#include "llvm/ExecutionEngine/JIT.h"
#include "llvm/Support/IRBuilder.h"
#include "llvm/Module.h"
#include "llvm/PassManager.h"
#include "llvm/Intrinsics.h"
#include "llvm/Analysis/Verifier.h"
#include "llvm/Target/TargetOptions.h"
#include "llvm/Transforms/Scalar.h"
#include "llvm/Support/Dwarf.h"
#include "llvm/Support/TargetSelect.h"
#include <string>
#include <sstream>
#include <stdexcept>

    namespace Compiler {
 
        namespace Detail {
            
            // copy pasting https://llvm.org/viewvc/llvm-project/llvm/trunk/examples/ExceptionDemo/ExceptionDemo.cpp?view=markup&sortby=file
            
            namespace Exceptions {
                
                extern "C" {

                typedef enum {
                  _URC_NO_REASON = 0,
                  _URC_FOREIGN_EXCEPTION_CAUGHT = 1,
                  _URC_FATAL_PHASE2_ERROR = 2,
                  _URC_FATAL_PHASE1_ERROR = 3,
                  _URC_NORMAL_STOP = 4,
                  _URC_END_OF_STACK = 5,
                  _URC_HANDLER_FOUND = 6,
                  _URC_INSTALL_CONTEXT = 7,
                  _URC_CONTINUE_UNWIND = 8
                } _Unwind_Reason_Code;

                typedef enum {
                  _UA_SEARCH_PHASE = 1,
                  _UA_CLEANUP_PHASE = 2,
                  _UA_HANDLER_FRAME = 4,
                  _UA_FORCE_UNWIND = 8,
                  _UA_END_OF_STACK = 16
                } _Unwind_Action;

                struct _Unwind_Exception;

                typedef void (*_Unwind_Exception_Cleanup_Fn) (_Unwind_Reason_Code,
                                                              struct _Unwind_Exception *);

                struct _Unwind_Exception {
                  uint64_t exception_class;
                  _Unwind_Exception_Cleanup_Fn exception_cleanup;

                  uintptr_t private_1;
                  uintptr_t private_2;

                  // @@@ The IA-64 ABI says that this structure must be double-word aligned.
                  //  Taking that literally does not make much sense generically.  Instead
                  //  we provide the maximum alignment required by any type for the machine.
                } __attribute__((__aligned__));
                
                  struct _Unwind_Context;
                typedef struct _Unwind_Context *_Unwind_Context_t;

                extern const uint8_t *_Unwind_GetLanguageSpecificData (_Unwind_Context_t c);
                extern uintptr_t _Unwind_GetGR (_Unwind_Context_t c, int i);
                extern void _Unwind_SetGR (_Unwind_Context_t c, int i, uintptr_t n);
                extern void _Unwind_SetIP (_Unwind_Context_t, uintptr_t new_value);
                extern uintptr_t _Unwind_GetIP (_Unwind_Context_t context);
                extern uintptr_t _Unwind_GetRegionStart (_Unwind_Context_t context);

              } // extern "C"
                
                /// This is our simplistic type info
                struct OurExceptionType_t {
                  /// type info type
                  int type;
                };
                
                /// This is our Exception class which relies on a negative offset to calculate
                /// pointers to its instances from pointers to its unwindException member.
                ///
                /// Note: The above unwind.h defines struct _Unwind_Exception to be aligned
                ///       on a double word boundary. This is necessary to match the standard:
                ///       http://refspecs.freestandards.org/abi-eh-1.21.html
                struct OurBaseException_t {
                  struct OurExceptionType_t type;

                  // Note: This is properly aligned in unwind.h
                  struct _Unwind_Exception unwindException;
                };
                
                // Note: Not needed since we are C++
                typedef struct OurBaseException_t OurException;
                typedef struct _Unwind_Exception OurUnwindException;

                //
                // Various globals used to support typeinfo and generatted exceptions in
                // general
                //

                static std::map<std::string, llvm::Value*> namedValues;

                inline int64_t& ourBaseFromUnwindOffset()
                {
                    static int64_t instance;
                    return instance;
                }

                const unsigned char ourBaseExcpClassChars[] =
                {'o', 'b', 'j', '\0', 'b', 'a', 's', '\0'};


                static uint64_t ourBaseExceptionClass = 0;

                static std::vector<std::string> ourTypeInfoNames;
                static std::map<int, std::string> ourTypeInfoNamesIndex;

                static llvm::StructType *ourTypeInfoType;
                static llvm::StructType *ourCaughtResultType;
                static llvm::StructType *ourExceptionType;
                static llvm::StructType *ourUnwindExceptionType;

                static llvm::ConstantInt *ourExceptionNotThrownState;
                static llvm::ConstantInt *ourExceptionThrownState;
                static llvm::ConstantInt *ourExceptionCaughtState;
                
                void initialiseExceptionHandlingStaticStuff(llvm::LLVMContext& context);

                typedef std::vector<std::string> ArgNames;
                typedef std::vector<llvm::Type*> ArgTypes;
                
                class OurCppRunException : public std::runtime_error {
                    public:
                    OurCppRunException(const std::string reason) :
                    std::runtime_error(reason) {}

                    OurCppRunException (const OurCppRunException &toCopy) :
                    std::runtime_error(toCopy) {}

                    OurCppRunException &operator = (const OurCppRunException &toCopy) {
                      return(reinterpret_cast<OurCppRunException&>(
                                                   std::runtime_error::operator=(toCopy)));
                    }

                    ~OurCppRunException (void) throw () {}
                  };
                
                typedef void (*OurExceptionThrowFunctType) (int32_t typeToThrow);
                
                typedef llvm::ArrayRef<llvm::Type*> TypeArray;
                
                /// Utility used to create a function, both declarations and definitions
                /// @param module for module instance
                /// @param retType function return type
                /// @param theArgTypes function's ordered argument types
                /// @param theArgNames function's ordered arguments needed if use of this
                ///        function corresponds to a function definition. Use empty
                ///        aggregate for function declarations.
                /// @param functName function name
                /// @param linkage function linkage
                /// @param declarationOnly for function declarations
                /// @param isVarArg function uses vararg arguments
                /// @returns function instance
                llvm::Function *createFunction(llvm::Module &module,
                                               llvm::Type *retType,
                                               const ArgTypes &theArgTypes,
                                               const ArgNames &theArgNames,
                                               const std::string &functName,
                                               llvm::GlobalValue::LinkageTypes linkage,
                                               bool declarationOnly,
                                               bool isVarArg);
                
                /// Create an alloca instruction in the entry block of
                /// the parent function.  This is used for mutable variables etc.
                /// @param function parent instance
                /// @param varName stack variable name
                /// @param type stack variable type
                /// @param initWith optional constant initialization value
                /// @returns AllocaInst instance
                llvm::AllocaInst *createEntryBlockAlloca(llvm::Function &function,
                                                                const std::string &varName,
                                                                llvm::Type *type,
                                                                llvm::Constant *initWith = 0);

                
                /// This initialization routine creates type info globals and
               /// adds external function declarations to module.
               /// @param numTypeInfos number of linear type info associated type info types
               ///        to create as GlobalVariable instances, starting with the value 1.
               /// @param module code for module instance
               void createStandardUtilityFunctions(unsigned numTypeInfos,
                                                          llvm::Module &module);
               
               
               void deleteFromUnwindOurException(_Unwind_Reason_Code reason,
                                  OurUnwindException *expToDelete);
               
               OurUnwindException *createOurException(int type);
               
               /// This is the personality function which is embedded (dwarf emitted), in the
                /// dwarf unwind info block. Again see: JITDwarfEmitter.cpp.
                /// See @link http://refspecs.freestandards.org/abi-eh-1.21.html @unlink
                /// @param version unsupported (ignored), unwind version
                /// @param _Unwind_Action actions minimally supported unwind stage
                ///        (forced specifically not supported)
                /// @param exceptionClass exception class (_Unwind_Exception::exception_class)
                ///        of thrown exception.
                /// @param exceptionObject thrown _Unwind_Exception instance.
                /// @param context unwind system context
                /// @returns minimally supported unwinding control indicator
                _Unwind_Reason_Code ourPersonality(int version,
                                                   _Unwind_Action actions,
                                                   uint64_t exceptionClass,
                                                   struct _Unwind_Exception *exceptionObject,
                                                   _Unwind_Context_t context);
                
                
                /// Generates code to handle finally block type semantics: always runs
                /// regardless of whether a thrown exception is passing through or the
                /// parent function is simply exiting. In addition to printing some state
                /// to stderr, this code will resume the exception handling--runs the
                /// unwind resume block, if the exception has not been previously caught
                /// by a catch clause, and will otherwise execute the end block (terminator
                /// block). In addition this function creates the corresponding function's
                /// stack storage for the exception pointer and catch flag status.
                /// @param context llvm context
                /// @param module code for module instance
                /// @param builder builder instance
                /// @param toAddTo parent function to add block to
                /// @param blockName block name of new "finally" block.
                /// @param functionId output id used for printing
                /// @param terminatorBlock terminator "end" block
                /// @param unwindResumeBlock unwind resume block
                /// @param exceptionCaughtFlag reference exception caught/thrown status storage
                /// @param exceptionStorage reference to exception pointer storage
                /// @param caughtResultStorage reference to landingpad result storage
                /// @returns newly created block
                static llvm::BasicBlock *createFinallyBlock(llvm::LLVMContext &context,
                                                            llvm::Module &module,
                        llvm::Module& exceptionModule,
                                                            llvm::IRBuilder<> &builder,
                                                            llvm::Function &toAddTo,
                                                            std::string &blockName,
                                                            std::string &functionId,
                                                            llvm::BasicBlock &terminatorBlock,
                                                            llvm::BasicBlock &unwindResumeBlock,
                                                            llvm::Value **exceptionCaughtFlag,
                                                            llvm::Value **exceptionStorage,
                                                            llvm::Value **caughtResultStorage);
                
                
                /// Generates catch block semantics which print a string to indicate type of
                /// catch executed, sets an exception caught flag, and executes passed in
                /// end block (terminator block).
                /// @param context llvm context
                /// @param module code for module instance
                /// @param builder builder instance
                /// @param toAddTo parent function to add block to
                /// @param blockName block name of new "catch" block.
                /// @param functionId output id used for printing
                /// @param terminatorBlock terminator "end" block
                /// @param exceptionCaughtFlag exception caught/thrown status
                /// @returns newly created block
                static llvm::BasicBlock* createCatchBlock(llvm::LLVMContext& context,
                                                          llvm::Module& module,
                        llvm::Module& exceptionModule,
                                                          llvm::IRBuilder<>& builder,
                                                          llvm::Function& toAddTo,
                                                          std::string& blockName,
                                                          std::string& functionId,
                                                          llvm::BasicBlock& terminatorBlock,
                                                          llvm::Value& exceptionCaughtFlag);
                
                
                /// Generates function which throws either an exception matched to a runtime
                /// determined type info type (argument to generated function), or if this
                /// runtime value matches nativeThrowType, throws a foreign exception by
                /// calling nativeThrowFunct.
                /// @param module code for module instance
                /// @param builder builder instance
                /// @param fpm a function pass manager holding optional IR to IR
                ///        transformations
                /// @param ourId id used to printing purposes
                /// @param nativeThrowType a runtime argument of this value results in
                ///        nativeThrowFunct being called to generate/throw exception.
                /// @param nativeThrowFunct function which will throw a foreign exception
                ///        if the above nativeThrowType matches generated function's arg.
                /// @returns generated function
                llvm::Function* createThrowExceptionFunction(llvm::Module &exceptionModule,
                                                             llvm::Module& insertModule,
                                                             llvm::IRBuilder<> &builder,
                                                             llvm::FunctionPassManager &fpm,
                                                             std::string ourId,
                                                             int32_t nativeThrowType,
                                                             llvm::Function &nativeThrowFunct);
                
                
                
                /// Generates a function which invokes a function (toInvoke) and, whose
                /// unwind block will "catch" the type info types correspondingly held in the
                /// exceptionTypesToCatch argument. If the toInvoke function throws an
                /// exception which does not match any type info types contained in
                /// exceptionTypesToCatch, the generated code will call _Unwind_Resume
                /// with the raised exception. On the other hand the generated code will
                /// normally exit if the toInvoke function does not throw an exception.
                /// The generated "finally" block is always run regardless of the cause of
                /// the generated function exit.
                /// The generated function is returned after being verified.
                /// @param module code for module instance
                /// @param builder builder instance
                /// @param fpm a function pass manager holding optional IR to IR
                ///        transformations
                /// @param toInvoke inner function to invoke
                /// @param ourId id used to printing purposes
                /// @param numExceptionsToCatch length of exceptionTypesToCatch array
                /// @param exceptionTypesToCatch array of type info types to "catch"
                /// @returns generated function
                struct InvokeFunctionResult
                {
                    llvm::Function* function;
                    llvm::InvokeInst* invokeInst;
                };
                
                InvokeFunctionResult createCatchWrappedInvokeFunction(llvm::Module& module,
                        llvm::Module &exceptionModule,
                                                             llvm::IRBuilder<>& builder,
                                                             llvm::FunctionPassManager& fpm,
                                                             llvm::Function& toInvoke,
                                                             std::string ourId,
                                                             unsigned numExceptionsToCatch,
                                                             unsigned exceptionTypesToCatch[]);
                
                
                /// Creates test code by generating and organizing these functions into the
                /// test case. The test case consists of an outer function setup to invoke
                /// an inner function within an environment having multiple catch and single
                /// finally blocks. This inner function is also setup to invoke a throw
                /// function within an evironment similar in nature to the outer function's
                /// catch and finally blocks. Each of these two functions catch mutually
                /// exclusive subsets (even or odd) of the type info types configured
                /// for this this. All generated functions have a runtime argument which
                /// holds a type info type to throw that each function takes and passes it
                /// to the inner one if such a inner function exists. This type info type is
                /// looked at by the generated throw function to see whether or not it should
                /// throw a generated exception with the same type info type, or instead call
                /// a supplied a function which in turn will throw a foreign exception.
                /// @param module code for module instance
                /// @param builder builder instance
                /// @param fpm a function pass manager holding optional IR to IR
                ///        transformations
                /// @param nativeThrowFunctName name of external function which will throw
                ///        a foreign exception
                /// @returns outermost generated test function.
                llvm::Function *createUnwindExceptionTest(llvm::Module &module,
                        llvm::Module& exceptionModule,
                                                          llvm::IRBuilder<> &builder,
                                                          llvm::FunctionPassManager &fpm,
                                                          std::string nativeThrowFunctName);
            }
        }
    }

#endif	/* EXCEPTIONHANDLING_H */

